from .ascim import ASCIM


class ASCIMDraw:
    """Draw text and shapes on an ASCIM Image. Operations will be done in-place."""

    def __init__(self, im):
        """Construct a ASCIMDraw object on a canvas.
        
        :param im: ASCIM Image this ASCIMDraw object applies modifications to.
        """
        
        if hasattr(im, 'size'):
            # HACK: very naïve duck type checking
            self.__image = im
        else:
            raise TypeError('`ASCIMDraw` requires an ASCIM Image as argument.')

    def text(self, box: tuple, text: str, transparency=False):
        """Draw text on ASCIM Image.

        :param box: box to fit the text in. (left, top, width, height)
        :param text: text to draw on ASCIM Image.
        :param transparency: if set to True, when the ``text`` to be drawn overlaps
            with non-whitespace characters in the original image, the character
            from the latter is kept. Otherwise, the character is a space.
        """

        wrapped = ''
        words = text.split(' ')
        width = 0
        for wd in words:
            if width + len(wd) <= box[2]:
                wrapped += wd + ' '
                width += len(wd) + 1
            else:
                wrapped += '\n' + wd + ' '
                width = len(wd) + 1
        # remove trailing space on each line, incl. last line
        wrapped = wrapped.replace(' \n', '\n')[:-1]
        self.__image.paste(ASCIM(wrapped), (box[0], box[1]))



from .utils import *


class ASCIM:
    """Basic type for an ASCIM Image."""

    def __init__(self, text: str):
        """Initialize an ASCIM Image.

        In an ASCIM Image, a character is a pixel,
        line (row) #A corresponds to y=A, and column #A, x=A.
        A is zero-indexed, just like real bitmap images.

        When line widths vary, ASCIM will pad spaces to lines so that
        line widths are uniform.

        :param text: string to transform to an ASCIM object.
        """

        # all of what are assigned to `self` here are private.
        rows = text.split('\n')
        self.__size = (max([len(l) for l in rows]), len(rows))
        for i, l in enumerate(rows):
            # fill up margins so all rows are uniform in length
            rows[i] += ' ' * (self.__size[0] - len(l))
        self.__rows = rows

    @classmethod
    def from_list(cls, ls: list):
        """Construct an ASCIM Image from a list of strings.

        :param ls: list from which ASCIM Image is constructed.
        """
        return ASCIM('\n'.join(ls))  # HACK

    @classmethod
    def new(cls, size: tuple, fill=' '):
        """Construct a blank new ASCIM Image with a repeating single character.

        :param size: tuple (width, height) as size of the new ASCIM Image.
        :param fill: character to fill the image with.
        """
        if not is_char(fill):
            raise ValueError('`fill` must be an ascii character')
        # [:-1]: remove trailing \n
        return ASCIM(((fill * size[0] + '\n') * size[1])[:-1])

    def copy(self):
        """:returns: A copy of image."""
        return ASCIM(self.to_text())

    def to_text(self) -> str:
        """:returns: Textual representation of image"""
        return '\n'.join(self.__rows)

    def show(self):
        """Prints image to stdout"""
        print(self.to_text())

    def to_rows(self) -> list:
        """:returns: A representation of image in a list of rows."""

        return self.__rows

    def to_columns(self) -> list:
        """:returns: A representation of image in a list of columns."""

        columns = []
        for x in range(self.__size[0]):
            col = ''
            for y in range(self.__size[1]):
                col += self.__rows[y][x]
            columns.append(col)
        return columns

    def char_at(self, x: int, y: int) -> str:
        """Retrieve a single character from image.

        :returns: If (x, y) is inside of bound, the character at (x, y);
            else, ``None``.
        """

        if self.__in_bound(x, y):
            return self.__rows[y][x]
        else:
            return None

    def set_char(self, x: int, y: int, char: str):
        """Set one character of self to `char` in-place.

        :raises: ValueError if `char` is not 1 in length.
        """

        if not isinstance(char, str) or not len(char) == 1:
            raise ValueError

        row = self.__rows[y]
        self.__rows[y] = row[:x] + char + row[x+1:]

    @property  # read-only
    def size(self) -> tuple:
        """:returns: size of image in (width, height) characters"""

        return self.__size

    def crop(self, box: tuple):
        """:returns: a cropped copy of an ASCIM Image.

        When any parameter in `box` is out of bound, ASCIM will use the closest
        edge instead. For example, when `self` is 33 characters wide (right edge
        x=32) but a crop to right edge x=36, ASCIM will make the right edge x=32.

        :param box: a tuple of (left, top, width, height).
        """

        return ASCIM.from_list(
            [r[box[0]:box[0]+box[2]] for r in self.__rows[box[1]:box[1]+box[3]]])

    def autocrop(self):
        """:returns: a copy of `self` with spaces on the edges cropped out. """

        copy = self.copy()
        while copy.to_columns()[0].isspace():  # left
            copy = copy.crop((1, 0, copy.size[0] - 1, copy.size[1]))
        while copy.to_columns()[-1].isspace():  # right
            copy = copy.crop((0, 0, copy.size[0] - 1, copy.size[1]))
        while copy.to_rows()[0].isspace():  # top
            copy = copy.crop((0, 1, copy.size[0], copy.size[1] - 1))
        while copy.to_rows()[-1].isspace():  # bottom
            copy = copy.crop((0, 0, copy.size[0], copy.size[1] - 1))

        return copy

    def paste(self, im, xy: tuple, transparency=''):
        """Pastes ASCIM Image `im` onto self in-place.

        When the location specified in `xy` exceeds image boundary, make no
        modification to `self`.

        :param im: ASCIM Image to paste on `self`.
        :param xy: a tuple of (left, top) for the corner to paste `im`.
        :param transparency: the character which, when in `im`, is interpreted as
            a transparent pixel and will not overwrite the corresponding character
            on the destination image. When set to a False value, no transparency
            is applied.
        """

        imx, imy = im.size
        boxx, boxy = xy

        # when in doubt, use brute force
        for y in range(self.__size[1]):
            if boxy <= y < boxy + imy:
                for x in range(self.__size[0]):
                    if (boxx <= x < boxx + imx and
                            not im.char_at(x - boxx, y - boxy) == transparency):
                        self.set_char(x, y, im.char_at(x - boxx, y - boxy))

    def __in_bound(self, x, y) -> bool:
        """Check if point(x, y) is inside image bound."""

        if 0 <= x < self.__size[0] and 0 <= y < self.__size[1]:
            return True
        return False



from setuptools import setup, find_packages

setup(
    name='ascim',
    version='0.1.1',
    description='Manipulate ASCII Art & Tables in a bitmap-image-like manner.',
    long_description=open('./README.md').read(),
    long_description_content_type='text/markdown',
    keywords='ascii art',
    license='MIT',
    url='https://github.com/fakefred/ascim',
    author='fakefred',
    author_email='fakefred@protonmail.ch',
    classifiers=[
        'Development Status :: 3 - Alpha',
        'License :: OSI Approved :: MIT License'
    ],
    packages=find_packages(),
    include_package_data=True,
    python_requires='>=3.5',
    install_requires=[],
    project_urls={
        'LiberaPay': 'https://liberapay.com/fakefred/donate'
    }
)



from .ascim import ASCIM
import re


class ASCIMTable:
    """ASCII Tables with simple manipulation functionalities.

    Each cell is an ASCIM Image, contrary to plaintext, although ASCIM Images
    can hold plaintext.

    If you're looking for a library that dumps csv or markdown, please look
    somewhere else.
    """

    def __init__(self, data: list):
        """Construct an ASCIM table object from a 2-dimensional list

        When rows are different in width, extra ``None`` are padded to their
        right until all rows have the same amount of columns.

        :param data: for a w*h table, this is a len=h list of len-w lists, each
            element of which populated by an ASCIM Image.
        """
        self.__size = (max([len(r) for r in data]), len(data))
        for i, r in enumerate(data):
            data[i].extend([None] * (self.__size[0] - len(r)))

        self.__data = data

    @classmethod
    def from_text(self, text: str):
        """Constructs an ASCIM Table from a string.

        Note that the string ``text`` must conform to a set of rules,
        as is generated from ASCIMTable's ``to_text()``.
        Valid table string:

        .. code-block::

            +-----+-----+------+
            | qwe | rty | uiop |
            +-----+-----+------+
            | asd | fgh | jkl  |
            +-----+-----+------+
            | zx  | cv  | bnm  |
            +-----+-----+------+

        """

        im = ASCIM(text)

        horizontal_lines = []
        vertical_lines = []

        for n, row in enumerate(im.to_rows()):
            if re.match('^\+(-+\+)+$', row):
                horizontal_lines.append(n)

        for n, col in enumerate(im.to_columns()):
            if re.match('^\+(\|+\+)+$', col):
                vertical_lines.append(n)

        if len(horizontal_lines) < 2 or len(horizontal_lines) < 2:
            raise ValueError('Invalid string to convert to table')

        cells = []
        for y in range(len(horizontal_lines) - 1):
            row = []
            y_start = horizontal_lines[y] + 1
            for x in range(len(vertical_lines) - 1):
                x_start = vertical_lines[x] + 2
                row.append(im.crop((
                    x_start, y_start,
                    vertical_lines[x + 1] - vertical_lines[x] - 3,
                    horizontal_lines[y + 1] - horizontal_lines[y] - 1
                )).autocrop())
            cells.append(row)

        return ASCIMTable(cells)

    def to_text(self) -> str:
        """:returns: textual representation of table."""
        # replace all None cells with an empty 1x1 ASCIM Image
        data = [[ASCIM.new((1, 1)) if col is None else col for col in row]
                for row in self.__data]
        table = ASCIMTable(data)
        # determine size of cells
        sizes = [[col.size for col in row] for row in data]
        widths = [max([row[x][0] for row in sizes])
                  for x in range(table.size[0])]
        heights = [max([col[1] for col in row]) for row in sizes]
        text = ''
        hl = '+'  # a line of +/- that make up horizontal lines
        vl = '|'  # a line of |/<space> that make up vertical lines
        for wd in widths:
            vl += ' ' * (wd + 2) + '|'
            hl += '-' * (wd + 2) + '+'
        hl += '\n'
        vl += '\n'
        for ht in heights:
            text += hl + vl * ht
        text += hl

        im = ASCIM(text)
        for rn, row in enumerate(data):
            for cn, col in enumerate(row):
                im.paste(col, (sum(widths[:cn]) + 3 * cn + 2,
                               sum(heights[:rn]) + rn + 1))

        return im.to_text()

    @classmethod
    def new(cls, size: tuple):
        """Construct a blank ASCIM Table of a certain size.

        :param size: tuple (width, height) as size of the new ASCIM Table,
            in number of cells.
        """

        return ASCIMTable([[None] * size[0]] * size[1])

    def copy(self):
        """:returns: a copy of `self`."""
        return ASCIMTable(self.__data)

    def cell_at(self, x: int, y: int):
        """Retrieve a single cell from table.

        :returns: ASCIM image or ``None`` as in the cell, or if ``(x, y)``
            exceeds bound, ``None``.
        """

        return self.__data[y][x]

    # TODO
    # def set_cell(self, x: int, y: int, cell):

    @property
    def size(self) -> tuple:
        """:returns: size of table in (width, height) cells"""

        return self.__size

    def __in_bound(self, x: int, y: int) -> bool:
        if 0 <= x < self.__size[0] and 0 <= y < self.__size[1]:
            return True
        return False
